@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2022, University of Amsterdam
+    Copyright (c)  1985-2024, University of Amsterdam
                               VU University Amsterdam
                               CWI, Amsterdam
                               SWI-Prolog Solutions b.v.
@@ -77,7 +77,6 @@
             rule/3,                             % :Head, -Rule, ?Ref
             numbervars/3,                       % +Term, +Start, -End
             term_string/3,                      % ?Term, ?String, +Options
-            nb_setval/2,                        % +Var, +Value
             thread_create/2,                    % :Goal, -Id
             thread_join/1,                      % +Id
             sig_block/1,                        % :Pattern
@@ -245,8 +244,8 @@ atom_prefix(Atom, Prefix) :-
 %   if the canonical name  as   defined  by  absolute_file_name/2 is
 %   known as a loaded filename.
 %
-%   Note that Time = 0.0 is used by  PlDoc and other code that needs
-%   to create a file record without being interested in the time.
+%   Note that Time = 0 is used by PlDoc and other code that needs to
+%   create a file record without being interested in the time.
 
 source_file(File) :-
     (   current_prolog_flag(access_level, user)
@@ -260,7 +259,7 @@ source_file(File) :-
         ), !
     ;   '$time_source_file'(File, Time, Level)
     ),
-    Time > 0.0.
+    float(Time).
 
 %!  source_file(+Head, -File) is semidet.
 %!  source_file(?Head, ?File) is nondet.
@@ -274,8 +273,7 @@ source_file(M:Head, File) :-
     !,
     (   '$c_current_predicate'(_, M:Head),
         predicate_property(M:Head, multifile)
-    ->  multi_source_files(M:Head, Files),
-        '$member'(File, Files)
+    ->  multi_source_file(M:Head, File)
     ;   '$source_file'(M:Head, File)
     ).
 source_file(M:Head, File) :-
@@ -286,18 +284,15 @@ source_file(M:Head, File) :-
     '$source_file_predicates'(File, Predicates),
     '$member'(M:Head, Predicates).
 
-:- thread_local found_src_file/1.
-
-multi_source_files(Head, Files) :-
-    call_cleanup(
-        findall(File, multi_source_file(Head, File), Files),
-        retractall(found_src_file(_))).
-
 multi_source_file(Head, File) :-
+    State = state([]),
     nth_clause(Head, _, Clause),
     clause_property(Clause, source(File)),
-    \+ found_src_file(File),
-    asserta(found_src_file(File)).
+    arg(1, State, Found),
+    (   memberchk(File, Found)
+    ->  fail
+    ;   nb_linkarg(1, State, [File|Found])
+    ).
 
 
 %!  source_file_property(?File, ?Property) is nondet.
@@ -1412,19 +1407,6 @@ term_string(Term, String, Options) :-
     format(string(String), '~W', [Term, Options1]).
 
 
-                 /*******************************
-                 *             GVAR             *
-                 *******************************/
-
-%!  nb_setval(+Name, +Value) is det.
-%
-%   Bind the non-backtrackable variable Name with a copy of Value
-
-nb_setval(Name, Value) :-
-    duplicate_term(Value, Copy),
-    nb_linkval(Name, Copy).
-
-
 		 /*******************************
 		 *            THREADS		*
 		 *******************************/
@@ -1491,8 +1473,7 @@ unblock([H|T], P, List) :-
 
 signal_is_blocked(Head) :-
     nb_current('$sig_blocked', List),
-    '$member'(Head, List),
-    !.
+    memberchk(Head, List).
 
 %!  set_prolog_gc_thread(+Status)
 %
@@ -1617,8 +1598,11 @@ run_undo([H|T], E0, E) :-
     ;   '$domain_error'(most_general_term, Head)
     ),
     atomic_list_concat(['$wrap$', PName], WrapName),
-    volatile(M:WrapName/Arity),
-    module_transparent(M:WrapName/Arity),
+    PI = M:WrapName/Arity,
+    dynamic(PI),
+    '$notransact'(PI),
+    volatile(PI),
+    module_transparent(PI),
     WHead =.. [WrapName|Args],
     '$c_wrap_predicate'(M:Head, WName, Closure, Wrapped, M:(WHead :- Body)).
 
