@@ -107,7 +107,7 @@ typedef void *datum;
 
 typedef struct atom_chash
 { size_t allocated;
-  datum  atoms[1];
+  datum  atoms[];
 } atom_chash;
 
 #define SIZEOF_ATOM_HASH(n)	offsetof(atom_chash, atoms[n])
@@ -272,7 +272,7 @@ atom_to_datum(atom_t a)
 
 static inline datum
 integer_to_datum(intptr_t v)
-{ return (datum)(v<<1);
+{ return (datum)((uintptr_t)v<<1);
 }
 
 
