@@ -1,9 +1,10 @@
 /*  Part of XPCE --- The SWI-Prolog GUI toolkit
 
     Author:        Jan Wielemaker and Anjo Anjewierden
-    E-mail:        wielemak@science.uva.nl
-    WWW:           http://www.swi.psy.uva.nl/projects/xpce/
-    Copyright (c)  1985-2007, University of Amsterdam
+    E-mail:        jan@swi-prolog.org
+    WWW:           https://www.swi-prolog.org/
+    Copyright (c)  1985-2025, University of Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -34,6 +35,16 @@
 
 #define PCE_INCLUDED 1
 
+#ifdef _MSC_VER
+#define _CRT_SECURE_NO_WARNINGS
+#pragma warning(disable : 4996)	/* deprecate open() etc */
+#pragma warning(disable : 4244)	/* possible loss of data (int conversion) */
+#pragma warning(disable : 4267)	/* possible loss of data (int conversion) */
+#pragma warning(disable : 4018)	/* signed/unsigned comparison */
+#pragma warning(disable : 4146)	/* unary minus operator applied to unsigned type */
+#include <inttypes.h>
+#endif
+
 #include <md.h>
 
 #define O_NOX11RESOURCES 1		/* use own resource parser */
@@ -273,14 +284,6 @@ cpdata(to, from, type, n)
 
 #if defined(__unix__) && !defined(unix)
 #define unix 1
-#endif
-
-		 /*******************************
-		 *	 LIBRARY PROBLEMS	*
-		 *******************************/
-
-#ifndef StrTod
-#define StrTod(s, e)	strtod(s, e)
 #endif
 
 		/********************************
@@ -472,8 +475,8 @@ test, conversion and computation macro's are provided.
 
 #define LocalString(name, iswide, size) \
   string _s_ ## name ## _hdr; \
-  void  *_s_ ## name ## _buf = (void *)alloca(iswide ? (size) * sizeof(charW) \
-						   : (size) * sizeof(charA)); \
+  void  *_s_ ## name ## _buf = alloca((iswide) ? (size) * sizeof(charW)	\
+					       : (size) * sizeof(charA)); \
   PceString name = fstr_inithdr(&_s_ ## name ## _hdr, iswide, _s_ ## name ## _buf, size)
 
 		/********************************
@@ -678,7 +681,7 @@ subclass of class `object'.
 
 struct instance
 { OBJECT_HEADER				/* non-pce header */
-  Any		slots[1];		/* array of slots. */
+  Any		slots[];		/* array of slots. */
 };
 
 NewClass(object)
