@@ -3,7 +3,8 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2013-2015, VU University Amsterdam
+    Copyright (c)  2013-2022, VU University Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -107,13 +108,13 @@ init_locale_strings(PL_locale *l, struct lconv *conv)
     l->thousands_sep = ls_to_wcs(conv->thousands_sep, L",");
     l->grouping      = strdup(conv->grouping);
 
-    return TRUE;
+    return true;
   } else
   { l->decimal_point = wcsdup(L".");
     l->thousands_sep = wcsdup(L",");
     l->grouping      = strdup("\003");
 
-    return FALSE;
+    return false;
   }
 }
 
@@ -165,6 +166,18 @@ update_locale(PL_locale *l, int category, const char *locale)
 }
 
 
+static void
+free_locale_symbol(table_key_t name, table_value_t value)
+{ PL_locale *l = val2ptr(value);
+  atom_t alias = word2atom(name);
+
+  l->alias = 0;
+  PL_unregister_atom(alias);
+
+  releaseLocale(l);
+}
+
+
 static int
 alias_locale(PL_locale *l, atom_t alias)
 { GET_LD
@@ -173,21 +186,23 @@ alias_locale(PL_locale *l, atom_t alias)
   PL_LOCK(L_LOCALE);
 
   if ( !GD->locale.localeTable )
-    GD->locale.localeTable = newHTable(16);
+  { GD->locale.localeTable = newHTableWP(16);
+    GD->locale.localeTable->free_symbol = free_locale_symbol;
+  }
 
-  if ( lookupHTable(GD->locale.localeTable, (void*)alias) )
+  if ( lookupHTableWP(GD->locale.localeTable, alias) )
   { GET_LD
     term_t obj = PL_new_term_ref();
 
     PL_put_atom(obj, alias);
     rc = PL_error("locale_create", 2, "Alias name already taken",
 		  ERR_PERMISSION, ATOM_create, ATOM_locale, obj);
-  }
-  else
-  { addNewHTable(GD->locale.localeTable, (void*)alias, l);
+  } else
+  { addNewHTableWP(GD->locale.localeTable, alias, l);
     l->alias = alias;
     PL_register_atom(alias);
-    rc = TRUE;
+    l->references++;	/* acquireLocale(), but that will deadlock */
+    rc = true;
   }
   PL_UNLOCK(L_LOCALE);
 
@@ -211,7 +226,7 @@ write_locale_ref(IOSTREAM *s, atom_t aref, int flags)
 
   Sfprintf(s, "<locale>(%p)", ref->data);
 
-  return TRUE;
+  return true;
 }
 
 
@@ -219,7 +234,7 @@ static void
 acquire_locale_ref(atom_t aref)
 { locale_ref *ref = PL_blob_data(aref, NULL, NULL);
 
-  (void)ref;
+  acquireLocale(ref->data);
 }
 
 
@@ -228,13 +243,13 @@ release_locale_ref(atom_t aref)
 { locale_ref *ref = PL_blob_data(aref, NULL, NULL);
 
   PL_LOCK(L_LOCALE);
-  if ( ref->data->references == 0 )
+  if ( --ref->data->references == 0 )
     free_locale(ref->data);
   else
     ref->data->symbol = 0;
   PL_UNLOCK(L_LOCALE);
 
-  return TRUE;
+  return true;
 }
 
 
@@ -290,7 +305,7 @@ unifyLocale(term_t t, PL_locale *l, int alias)
     return PL_unify(t, b);
   }
 
-  return FALSE;
+  return false;
 }
 
 
@@ -311,17 +326,17 @@ getLocale(term_t t, PL_locale **lp)
     } else if ( (ref=PL_blob_data(a, NULL, &bt)) && bt == &locale_blob )
     { l = ref->data;
     } else if ( GD->locale.localeTable )
-    { l = lookupHTable(GD->locale.localeTable, (void*)a);
+    { l = lookupHTableWP(GD->locale.localeTable, a);
     }
 
     if ( l )
     { assert(l->magic == LOCALE_MAGIC);
       *lp = acquireLocale(l);
-      return TRUE;
+      return true;
     }
   }
 
-  return FALSE;
+  return false;
 }
 
 
@@ -330,14 +345,14 @@ getLocaleEx(term_t t, PL_locale **lp)
 { GET_LD
 
   if ( getLocale(t, lp) )
-    return TRUE;
+    return true;
 
   if ( PL_is_atom(t) )
     PL_existence_error("locale", t);
   else
     PL_type_error("locale", t);
 
-  return FALSE;
+  return false;
 }
 
 
@@ -345,40 +360,60 @@ getLocaleEx(term_t t, PL_locale **lp)
 		 *	 PROLOG BINDING		*
 		 *******************************/
 
+#define locale_alias_property(l, prop) \
+	LDFUNC(locale_alias_property, l, prop)
+
 static int		/* locale_property(Mutex, alias(Name)) */
-locale_alias_property(PL_locale *l, term_t prop ARG_LD)
-{ if ( l->alias )
+locale_alias_property(DECL_LD void *ctx, term_t prop)
+{ PL_locale *l = ctx;
+
+  if ( l->alias )
     return PL_unify_atom(prop, l->alias);
 
-  return FALSE;
+  return false;
 }
 
+#define locale_decimal_point_property(l, prop)  \
+	LDFUNC(locale_decimal_point_property, l, prop)
+
 static int		/* locale_property(Locale, decimal_point(Atom)) */
-locale_decimal_point_property(PL_locale *l, term_t prop ARG_LD)
-{ if ( l->decimal_point && l->decimal_point[0] )
+locale_decimal_point_property(DECL_LD void *ctx, term_t prop)
+{ PL_locale *l = ctx;
+
+  if ( l->decimal_point && l->decimal_point[0] )
     return PL_unify_wchars(prop, PL_ATOM, (size_t)-1, l->decimal_point);
 
-  return FALSE;
+  return false;
 }
 
+#define locale_thousands_sep_property(l, prop) \
+	LDFUNC(locale_thousands_sep_property, l, prop)
+
 static int		/* locale_property(Locale, thousands_sep(Atom)) */
-locale_thousands_sep_property(PL_locale *l, term_t prop ARG_LD)
-{ if ( l->thousands_sep && l->thousands_sep[0] )
+locale_thousands_sep_property(DECL_LD void *ctx, term_t prop)
+{ PL_locale *l = ctx;
+
+  if ( l->thousands_sep && l->thousands_sep[0] )
     return PL_unify_wchars(prop, PL_ATOM, (size_t)-1, l->thousands_sep);
 
-  return FALSE;
+  return false;
 }
 
+#define locale_grouping_property(l, prop) \
+	LDFUNC(locale_grouping_property, l, prop)
+
 static int		/* locale_property(Locale, grouping(List)) */
-locale_grouping_property(PL_locale *l, term_t prop ARG_LD)
-{ if ( l->grouping && l->grouping[0] )
+locale_grouping_property(DECL_LD void *ctx, term_t prop)
+{ PL_locale *l = ctx;
+
+  if ( l->grouping && l->grouping[0] )
   { term_t tail = PL_copy_term_ref(prop);
     term_t head = PL_new_term_ref();
     char *s;
 
     for(s=l->grouping; ; s++)
     { if ( !PL_unify_list(tail, head, tail) )
-	return FALSE;
+	return false;
       if ( s[1] == 0 || (s[1] == s[0] && s[2] == 0) )
 	return ( PL_unify_term(head, PL_FUNCTOR, FUNCTOR_repeat1,
 			       PL_INT, (int)s[0]) &&
@@ -387,19 +422,19 @@ locale_grouping_property(PL_locale *l, term_t prop ARG_LD)
       if ( s[0] == CHAR_MAX )
 	return PL_unify_nil(tail);
       if ( !PL_unify_integer(head, s[0]) )
-	return FALSE;
+	return false;
     }
   }
 
-  return FALSE;
+  return false;
 }
 
 
 static const tprop lprop_list [] =
-{ { FUNCTOR_alias1,	    locale_alias_property },
-  { FUNCTOR_decimal_point1, locale_decimal_point_property },
-  { FUNCTOR_thousands_sep1, locale_thousands_sep_property },
-  { FUNCTOR_grouping1,      locale_grouping_property },
+{ { FUNCTOR_alias1,	    LDFUNC_REF(locale_alias_property) },
+  { FUNCTOR_decimal_point1, LDFUNC_REF(locale_decimal_point_property) },
+  { FUNCTOR_thousands_sep1, LDFUNC_REF(locale_thousands_sep_property) },
+  { FUNCTOR_grouping1,      LDFUNC_REF(locale_grouping_property) },
   { 0,			    NULL }
 };
 
@@ -408,6 +443,7 @@ typedef struct
   PL_locale *l;				/* current locale */
   const tprop *p;			/* Pointer in properties */
   int enum_properties;			/* Enumerate the properties */
+  int allocated;			/* Is allocated */
 } lprop_enum;
 
 
@@ -416,21 +452,22 @@ advance_lstate(lprop_enum *state)
 { if ( state->enum_properties )
   { state->p++;
     if ( state->p->functor )
-      return TRUE;
+      return true;
 
     state->p = lprop_list;
   }
   if ( state->e )
-  { PL_locale *l;
+  { table_value_t tv;
+    if ( advanceTableEnum(state->e, NULL, &tv) )
+    { PL_locale *l = val2ptr(tv);
 
-    if ( advanceTableEnum(state->e, NULL, (void**)&l) )
-    { state->l = l;
+      state->l = l;
 
-      return TRUE;
+      return true;
     }
   }
 
-  return FALSE;
+  return false;
 }
 
 
@@ -441,7 +478,8 @@ free_lstate(lprop_enum *state)
   else if ( state->l )
     releaseLocale(state->l);
 
-  freeForeignState(state, sizeof(*state));
+  if ( state->allocated )
+    freeForeignState(state, sizeof(*state));
 }
 
 
@@ -480,21 +518,21 @@ PRED_IMPL("locale_property", 2, locale_property, PL_FA_NONDETERMINISTIC)
 		 get_atom_arg(property, &alias) )
 	    { PL_locale *l;
 
-	      if ( (l = lookupHTable(GD->locale.localeTable, (void*)alias)) )
-		return unifyLocale(locale, l, FALSE);
+	      if ( (l = lookupHTableWP(GD->locale.localeTable, alias)) )
+		return unifyLocale(locale, l, false);
 	      else
-		return FALSE;
+		return false;
 	    }
-	    state->e = newTableEnum(GD->locale.localeTable);
+	    state->e = newTableEnumWP(GD->locale.localeTable);
 	    goto enumerate;
 	  }
 	  case 0:
-	    state->e = newTableEnum(GD->locale.localeTable);
+	    state->e = newTableEnumWP(GD->locale.localeTable);
 	    state->p = lprop_list;
-	    state->enum_properties = TRUE;
+	    state->enum_properties = true;
 	    goto enumerate;
 	  case -1:
-	    return FALSE;
+	    return false;
 	}
       } else if ( getLocale(locale, &state->l) )
       { switch( get_prop_def(property, ATOM_locale_property,
@@ -503,13 +541,14 @@ PRED_IMPL("locale_property", 2, locale_property, PL_FA_NONDETERMINISTIC)
 	    goto enumerate;
 	  case 0:
 	    state->p = lprop_list;
-	    state->enum_properties = TRUE;
+	    state->enum_properties = true;
 	    goto enumerate;
 	  case -1:
-	    return FALSE;
+	    free_lstate(state);
+	    return false;
 	}
       } else
-      { return FALSE;
+      { return false;
       }
     }
     case FRG_REDO:
@@ -526,15 +565,17 @@ PRED_IMPL("locale_property", 2, locale_property, PL_FA_NONDETERMINISTIC)
 
 enumerate:
   if ( !state->l )			/* first time, enumerating locales */
-  { PL_locale *l;
+  { table_value_t tv;
 
     assert(state->e);
-    if ( advanceTableEnum(state->e, NULL, (void**)&l) )
-    { state->l = l;
+    if ( advanceTableEnum(state->e, NULL, &tv) )
+    { PL_locale *l = val2ptr(tv);
+
+      state->l = l;
     } else
     { freeTableEnum(state->e);
       assert(state != &statebuf);
-      return FALSE;
+      return false;
     }
   }
 
@@ -544,39 +585,38 @@ enumerate:
       _PL_get_arg(1, property, arg);
 
     for(;;)
-    { if ( (*state->p->function)(state->l, arg PASS_LD) )
+    { if ( LDFUNCP(*state->p->function)(state->l, arg) )
       { if ( state->enum_properties )
 	{ if ( !PL_unify_term(property,
 			      PL_FUNCTOR, state->p->functor,
-			        PL_TERM, arg) )
+				PL_TERM, arg) )
 	    goto error;
 	}
 	if ( state->e )
-	{ if ( !unifyLocale(locale, state->l, TRUE) )
+	{ if ( !unifyLocale(locale, state->l, true) )
 	    goto error;
 	}
 
 	if ( advance_lstate(state) )
-	{ if ( state == &statebuf )
+	{ if ( !state->allocated )
 	  { lprop_enum *copy = allocForeignState(sizeof(*copy));
 
 	    *copy = *state;
+	    copy->allocated = true;
 	    state = copy;
 	  }
 
 	  ForeignRedoPtr(state);
 	}
 
-	if ( state != &statebuf )
-	  free_lstate(state);
-	return TRUE;
+	free_lstate(state);
+	return true;
       }
 
       if ( !advance_lstate(state) )
       { error:
-	if ( state != &statebuf )
-	  free_lstate(state);
-	return FALSE;
+	free_lstate(state);
+	return false;
       }
     }
   }
@@ -590,11 +630,11 @@ set_chars(term_t t, wchar_t **valp)
   if ( PL_get_wchars(t, NULL, &s, CVT_ATOM|CVT_EXCEPTION) )
   { free(*valp);
     if ( (*valp = wcsdup(s)) )
-      return TRUE;
+      return true;
     return PL_no_memory();
   }
 
-  return FALSE;
+  return false;
 }
 
 
@@ -607,12 +647,12 @@ get_group_size_ex(term_t t, int *s)
   if ( PL_get_integer_ex(t, &i) )
   { if ( i > 0 && i < CHAR_MAX )
     { *s = i;
-      return TRUE;
+      return true;
     }
     PL_domain_error("digit_group_size", t);
   }
 
-  return FALSE;
+  return false;
 }
 
 
@@ -632,19 +672,19 @@ set_grouping(term_t t, char **valp)
 
     if ( PL_is_functor(head, FUNCTOR_repeat1) )
     { if ( !PL_get_nil_ex(tail) )
-	return FALSE;
+	return false;
 
       _PL_get_arg(1, head, head);
       if ( get_group_size_ex(head, &g) )
-      { *o++ = g;
+      { *o++ = (char)g;
 	goto end;
       }
-      return FALSE;
+      return false;
     }
     if ( get_group_size_ex(head, &g) )
-    { *o++ = g;
+    { *o++ = (char)g;
     } else
-      return FALSE;
+      return false;
   }
 
   if ( PL_get_nil_ex(tail) )
@@ -653,11 +693,11 @@ set_grouping(term_t t, char **valp)
     *o++ = '\0';
     free(*valp);
     if ( (*valp = strdup(s)) )
-      return TRUE;
+      return true;
     return PL_no_memory();
   }
 
-  return FALSE;
+  return false;
 }
 
 
@@ -732,13 +772,15 @@ PRED_IMPL("locale_create", 3, locale_create, 0)
     {
     error:
       free_locale(new);
-      return FALSE;
+      return false;
     }
 
     if ( alias && !alias_locale(new, alias) )
       goto error;
 
-    return unifyLocale(A1, new, TRUE);
+    int rc = unifyLocale(A1, new, true);
+    DEBUG(0, assert(new->references == 1));
+    return rc;
   } else
   { return PL_no_memory();
   }
@@ -754,19 +796,18 @@ PRED_IMPL("locale_destroy", 1, locale_destroy, 0)
   { if ( l->alias )
     { atom_t alias = l->alias;
 
-      PL_LOCK(L_LOCALE);
-      if ( lookupHTable(GD->locale.localeTable, (void*)alias) )
-	deleteHTable(GD->locale.localeTable, (void*)alias);
-      l->alias = 0;
-      PL_unregister_atom(alias);
-      PL_UNLOCK(L_LOCALE);
+      if ( deleteHTableWP(GD->locale.localeTable, alias) == l )
+      { l->alias = 0;
+	l->references--;
+	PL_unregister_atom(alias);
+      }
     }
 
     releaseLocale(l);
-    return TRUE;
+    return true;
   }
 
-  return FALSE;
+  return false;
 }
 
 
@@ -796,10 +837,10 @@ PRED_IMPL("set_locale", 1, set_locale, 0)
       }
     }
 
-    return TRUE;
+    return true;
   }
 
-  return FALSE;
+  return false;
 }
 
 
@@ -811,9 +852,9 @@ PRED_IMPL("current_locale", 1, current_locale, 0)
 { PRED_LD
 
   if ( LD->locale.current )
-    return unifyLocale(A1, LD->locale.current, TRUE);
+    return unifyLocale(A1, LD->locale.current, true);
 
-  return FALSE;
+  return false;
 }
 
 
@@ -846,8 +887,7 @@ initLocale(void)
 
   if ( (def = new_locale(NULL)) )
   { alias_locale(def, ATOM_default);
-    def->references++;
-    GD->locale.default_locale = def;
+    GD->locale.default_locale = acquireLocale(def);
     LD->locale.current = acquireLocale(def);
 
     initDefaultsStreamsLocale(def);
@@ -874,7 +914,7 @@ initStreamLocale(IOSTREAM *s)
   if ( l )
     s->locale = acquireLocale(l);
 
-  return TRUE;
+  return true;
 }
 
 
@@ -891,12 +931,26 @@ acquireLocale(PL_locale *l)
 void
 releaseLocale(PL_locale *l)
 { PL_LOCK(L_LOCALE);
+  assert(l->references > 0);
   if ( --l->references == 0 && l->symbol == 0 && l->alias == 0 )
     free_locale(l);
   PL_UNLOCK(L_LOCALE);
 }
 
 
+void
+cleanupLocale(void)
+{ if ( GD->locale.default_locale )
+  { releaseLocale(GD->locale.default_locale);
+    GD->locale.default_locale = NULL;
+  }
+
+  if ( GD->locale.localeTable )
+  { destroyHTableWP(GD->locale.localeTable);
+    GD->locale.localeTable = NULL;
+  }
+}
+
 
 
 		 /*******************************
