@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2010-2015, University of Amsterdam
+    Copyright (c)  2010-2023, University of Amsterdam
                               VU University Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -35,10 +36,15 @@
 
 /*#define O_DEBUG 1*/
 #include "pl-incl.h"
+#include "pl-termhash.h"
+#include "pl-fli.h"
+#include "os/pl-text.h"
 #include "pl-arith.h"
 #include "pl-pro.h"
 #define AC_TERM_WALK 1
 #include "pl-termwalk.c"
+#undef LD
+#define LD LOCAL_LD
 
 #ifdef __WINDOWS__
 typedef unsigned int uint32_t;
@@ -64,41 +70,35 @@ typedef struct th_data
 } th_data;
 
 
-static void *
-allocBuffer(Buffer b, size_t size)
-{ void *ptr;
-
-  if ( b->top + size > b->max )
-  { if ( !growBuffer(b, size) )
-      return NULL;
-  }
-  ptr = b->top;
-  b->top += size;
-
-  return ptr;
-}
-
-
 #define primitiveHashValue(term, hval) LDFUNC(primitiveHashValue, term, hval)
 static int
 primitiveHashValue(DECL_LD word term, unsigned int *hval)
 { switch(tag(term))
   { case TAG_VAR:
     case TAG_ATTVAR:
-      return FALSE;
+      return false;
     case TAG_ATOM:
     { *hval = MurmurHashAligned2(&atomValue(term)->hash_value,
 				 sizeof(unsigned int), *hval);
-      return TRUE;
+      return true;
     }
     case TAG_STRING:
     { size_t len;
       char *s;
 
-      s = getCharsString(term, &len);
-      *hval = MurmurHashAligned2(s, len, *hval);
+      if ( (s = getCharsString(term, &len)) )
+      { *hval = MurmurHashAligned2(s, len, *hval);
+      } else
+      { PL_chars_t text;
+	if ( get_string_text(term, &text) &&
+	     PL_mb_text(&text, REP_UTF8) )
+	{ *hval = MurmurHashAligned2(text.text.t, text.length, *hval);
+	  PL_free_text(&text);
+	} else
+	  return false;
+      }
 
-      return TRUE;
+      return true;
     }
     case TAG_INTEGER:
       if ( storage(term) == STG_INLINE )
@@ -106,7 +106,7 @@ primitiveHashValue(DECL_LD word term, unsigned int *hval)
 
 	*hval = MurmurHashAligned2(&v, sizeof(v), *hval);
 
-	return TRUE;
+	return true;
       }
     /*FALLTHROUGH*/
     case TAG_FLOAT:
@@ -115,11 +115,11 @@ primitiveHashValue(DECL_LD word term, unsigned int *hval)
 
 	*hval = MurmurHashAligned2(p+1, n*sizeof(word), *hval);
 
-	return TRUE;
+	return true;
       }
     default:
       assert(0);
-      return FALSE;
+      return false;
   }
 }
 
@@ -130,7 +130,7 @@ start_term(DECL_LD th_data *work, Buffer b, word w)
 { atom_t name;
 
   work->term     = valueTerm(w);
-  work->functor  = work->term->definition;
+  work->functor  = word2functor(work->term->definition);
   work->hash     = MURMUR_SEED;
   work->arg      = 0;
   work->in_cycle = 0;
@@ -181,7 +181,7 @@ next_arg(DECL_LD th_data **workp, Buffer b)
       if ( work->in_cycle /*&& parent->functor == work->functor*/ ) /* (*) */
       { DEBUG(1, Sdprintf("Mark parent %ld as cycle\n",
 			  nodeID(parent, b)));
-	parent->in_cycle = TRUE;
+	parent->in_cycle = true;
       }
       DEBUG(1, Sdprintf("Updated hash in node %ld%s to %d\n",
 			work->parent_offset,
@@ -209,7 +209,7 @@ update_cycle(th_data *here, th_data *start, Buffer b)
   for(;;)
   { DEBUG(1, Sdprintf("Marking cycle for node %ld\n",
 		      nodeID(here, b)));
-    here->in_cycle = TRUE;
+    here->in_cycle = true;
     if ( here == start )
       break;
     here = nodePTR(here->parent_offset, b);
@@ -228,11 +228,11 @@ termHashValue(DECL_LD Word p, unsigned int *hval)
   { tmp_buffer tmp;
     Buffer b = (Buffer)&tmp;
     th_data *work;
-    int rc = TRUE;
+    int rc = true;
     Functor t = valueTerm(*p);
 
     initBuffer(&tmp);
-    work = allocBuffer(b, sizeof(*work));	/* cannot fail */
+    work = allocFromBuffer(b, sizeof(*work));	/* cannot fail */
     start_term(work, b, *p);
     work->parent_offset = (size_t)-1;
     t->definition = consInt(0);
@@ -242,7 +242,7 @@ termHashValue(DECL_LD Word p, unsigned int *hval)
       { if ( primitiveHashValue(*p, &work->hash) )
 	{ work->arg++;
 	} else
-	{ rc = FALSE;
+	{ rc = false;
 	  goto out;
 	}
       } else
@@ -262,13 +262,13 @@ termHashValue(DECL_LD Word p, unsigned int *hval)
 			      valInt(t->definition), seen->hash,
 			      nodeID(work, b), work->hash));
 	    if ( seen->in_cycle /*&& seen->functor == work->functor*/ )
-	      work->in_cycle = TRUE;
+	      work->in_cycle = true;
 	  }
 	  work->arg++;
 	} else
 	{ size_t parent = nodeID(work, b);
 
-	  if ( !(work = allocBuffer(b, sizeof(*work))) )
+	  if ( !(work = allocFromBuffer(b, sizeof(*work))) )
 	  { rc = -1;			/* out of memory */
 	    goto out;
 	  }
@@ -283,7 +283,7 @@ termHashValue(DECL_LD Word p, unsigned int *hval)
     { th_data *d   = baseBuffer(b, th_data);
       th_data *end = d + entriesBuffer(b, th_data);
 
-      if ( rc == TRUE )
+      if ( rc == true )
 	*hval = d->hash;
 
       for(; d<end; d++)
@@ -318,7 +318,7 @@ PRED_IMPL("term_hash", 2, term_hash, 0)
     return PL_unify_integer(A2, hraw);
   }
 
-  return TRUE;
+  return true;
 }
 
 
@@ -328,8 +328,6 @@ PRED_IMPL("term_hash", 2, term_hash, 0)
 
 /* type to hold the SHA256 context  */
 
-#define SHA1_DIGEST_SIZE 20
-
 typedef struct
 {   uint32_t count[2];
     uint32_t hash[5];
@@ -401,11 +399,6 @@ hash_end(hash_state *state)
 State for processing the term
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-typedef enum
-{ HASH_SHA1,
-  HASH_MURMUR
-} hash_algo;
-
 typedef struct
 { int		var_count;
   hash_algo	algorithm;
@@ -432,7 +425,7 @@ push_var(Word p, sha1_state *state)
 
 static int
 push_attvar(Word p, sha1_state *state)
-{ Word w = (Word)*p;
+{ Word w = word2ptr(Word, *p);
   return ( pushSegStack(&state->vars, p, Word) &&
 	   pushSegStack(&state->vars, w, Word) );
 }
@@ -496,7 +489,7 @@ variant_sha1(DECL_LD ac_term_agenda *agenda, sha1_state *state)
       }
       case TAG_INTEGER:
       { if ( !isIndirect(w) )
-	{ int64_t val = valInteger(w);
+	{ int64_t val = valInt(w);
 
 	  HASH("i", 1);
 	  HASH(&val, sizeof(val));
@@ -525,7 +518,7 @@ variant_sha1(DECL_LD ac_term_agenda *agenda, sha1_state *state)
 	switch(ac_pushTermAgenda(agenda, w, &f))
 	{ case -1:			/* Resource error */
 	    return E_RESOURCE;
-	  case FALSE:			/* Cycle */
+	  case false:			/* Cycle */
 	    return E_CYCLE;
 	  default:
 	  { FunctorDef fd = valueFunctor(f);
@@ -548,14 +541,12 @@ variant_sha1(DECL_LD ac_term_agenda *agenda, sha1_state *state)
 }
 
 
-#define variant_hash(term, hash, algorithm) LDFUNC(variant_hash, term, hash, algorithm)
-static int
-variant_hash(DECL_LD term_t term, term_t hash, hash_algo algorithm)
+int
+variant_hash(DECL_LD term_t term, termhash_t *hash, hash_algo algorithm)
 { int rc;
   ac_term_agenda agenda;
   sha1_state state;
   Word p;
-  int n;
 
   state.var_count = 0;
   state.algorithm = algorithm;
@@ -569,7 +560,7 @@ variant_hash(DECL_LD term_t term, term_t hash, hash_algo algorithm)
   rc = variant_sha1(&agenda, &state);
   ac_clearTermAgenda(&agenda);
   while(popSegStack(&state.vars, &p, Word))
-  { word w = (word)p;
+  { word w = ptr2word(p);
     if ( unlikely(isAttVar(w)) )
     { popSegStack(&state.vars, &p, Word);
       *p = w;
@@ -593,26 +584,11 @@ variant_hash(DECL_LD term_t term, term_t hash, hash_algo algorithm)
   }
 
   if ( state.algorithm == HASH_SHA1 )
-  { unsigned char sha1[SHA1_DIGEST_SIZE];
-    char hex[SHA1_DIGEST_SIZE*2];
-    const char hexd[] = "0123456789abcdef";
-    char *o;
-    const unsigned char *i;
+    sha1_end(hash->sha1, state.ctx.sha1);
+  else
+    hash->murmur = hash_end(state.ctx.murmur);
 
-    sha1_end(sha1, state.ctx.sha1);
-    o = hex;
-    i = sha1;
-    for(n=0; n<SHA1_DIGEST_SIZE; n++,i++)
-    { *o++ = hexd[*i >> 4];
-      *o++ = hexd[*i&0x0f];
-    }
-
-    return PL_unify_chars(hash, PL_ATOM|REP_ISO_LATIN_1, sizeof(hex), hex);
-  } else
-  { unsigned int key = hash_end(state.ctx.murmur)&PLMAXTAGGEDINT32;
-
-    return PL_unify_integer(hash, key);
-  }
+  return true;
 }
 
 /** variant_sha1(@Term, -SHA1:string) is det.
@@ -625,13 +601,38 @@ basically execute numbervars.
 static
 PRED_IMPL("variant_sha1", 2, variant_sha1, 0)
 { PRED_LD
-  return variant_hash(A1, A2, HASH_SHA1);
+  termhash_t hash;
+
+  if ( variant_hash(A1, &hash, HASH_SHA1) )
+  { char hex[SHA1_DIGEST_SIZE*2];
+    const char hexd[] = "0123456789abcdef";
+    char *o;
+    const unsigned char *i;
+    int n;
+
+    o = hex;
+    i = hash.sha1;
+    for(n=0; n<SHA1_DIGEST_SIZE; n++,i++)
+    { *o++ = hexd[*i >> 4];
+      *o++ = hexd[*i&0x0f];
+    }
+
+    return PL_unify_chars(A2, PL_ATOM|REP_ISO_LATIN_1, sizeof(hex), hex);
+  } else
+  { return false;
+  }
 }
 
 static
 PRED_IMPL("variant_hash", 2, variant_hash, 0)
 { PRED_LD
-  return variant_hash(A1, A2, HASH_MURMUR);
+  termhash_t hash;
+
+  if ( variant_hash(A1, &hash, HASH_MURMUR) )
+  { return PL_unify_integer(A2, hash.murmur&PLMAXTAGGEDINT32);
+  } else
+  { return false;
+  }
 }
 
 		 /*******************************
